/*
 * @(#)src/classes/sov/java/util/jar/JarVerifier.java, jar, as142 1.15.2.2
 * ===========================================================================
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 * IBM SDK, Java(tm) 2 Technology Edition, v1.4.2
 * (C) Copyright IBM Corp. 1998, 2004. All Rights Reserved
 * ===========================================================================
 */

/*
 * ===========================================================================
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * ===========================================================================
 */

 
/* 
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin    Description
 * ------  ------ --------  -------------------------------------------------- 
 * 063626  130803 pabbott   Improve debug tracing
 * 160250  260510 jitendsi  Integrating Sun Security Bulletin 6742654
 *                        
 */

package java.util.jar;

import java.io.*;
import java.net.URL;                                       /*ibm@160250*/
import java.security.CodeSource;                           /*ibm@160250*/
import java.security.NoSuchAlgorithmException;             /*ibm@160250*/
import java.security.SignatureException;                   /*ibm@160250*/
import java.security.cert.Certificate;                     /*ibm@160250*/
import java.security.cert.CertificateException;            /*ibm@160250*/
import java.util.*;
import java.util.zip.*;

import sun.security.util.ManifestDigester;
import sun.security.util.ManifestEntryVerifier;
import sun.security.util.SignatureFileVerifier;
import sun.security.util.Debug;

/**
 *
 * @version 	1.32 03/02/27
 * @author	Roland Schemers
 */
class JarVerifier {

    /* Are we debugging ? */
    static final Debug debug = Debug.getInstance("jar");

    /* a table mapping names to identities for entries that have
       had their actual hashes verified */
    private Hashtable verifiedCerts;

    /* a table mapping names to Certs for entries that have
       passed the .SF/.DSA -> MANIFEST check */
    private Hashtable sigFileCerts;

    /* a hash table to hold .SF bytes */
    private Hashtable sigFileData;

    /** "queue" of pending PKCS7 blocks that we couldn't parse
     *  until we parsed the .SF file */
    private ArrayList pendingBlocks;

    /* cache of Certificate[] objects */
    private ArrayList certCache;

    /* Are we parsing a block? */
    private boolean parsingBlockOrSF = false;

    /* Are we done parsing META-INF entries? */
    private boolean parsingMeta = true;

    /* Are there are files to verify? */
    private boolean anyToVerify = true;

    /* The manifest file */
    private Manifest manifest;

    /* The output stream to use when keeping track of files we are interested
       in */
    private ByteArrayOutputStream baos;

    /** The ManifestDigester object */
    private ManifestDigester manDig;

    /** the bytes for the manDig object */
    byte manifestRawBytes[] = null;

    /*ibm@160250-Starts*/
    /** controls eager signature validation */
    boolean eagerValidation;

    /** makes code source singleton instances unique to us */
    private Object csdomain;

    /** collect -DIGEST-MANIFEST values for blacklist */
    private List manifestDigests;
    /*ibm@160250-Ends*/

    public JarVerifier(Manifest manifest, byte rawBytes[]) {
        manifestRawBytes = rawBytes;
        sigFileCerts = new Hashtable();
        verifiedCerts = new Hashtable();
        sigFileData = new Hashtable(11);
        pendingBlocks = new ArrayList();
        baos = new ByteArrayOutputStream();
        manifestDigests = new ArrayList();                       /*ibm@160250*/
    }

    /**
     * This method scans to see which entry we're parsing and
     * keeps various state information depending on what type of
     * file is being parsed.
     */
    public void beginEntry(JarEntry je, ManifestEntryVerifier mev)
        throws IOException
    {
        if (je == null)
            return;

        if (debug != null) {
            debug.println("beginEntry "+je.getName());
        }

        String name = je.getName();

        /*
         * Assumptions:
         * 1. The manifest should be the first entry in the META-INF directory.
         * 2. The .SF/.DSA files follow the manifest, before any normal entries
         * 3. Any of the following will throw a SecurityException:
         *    a. digest mismatch between a manifest section and
         *       the SF section.
         *    b. digest mismatch between the actual jar entry and the manifest
         */

        if (parsingMeta) {
            String uname = name.toUpperCase(Locale.ENGLISH);
            if ((uname.startsWith("META-INF/") ||
                 uname.startsWith("/META-INF/"))) {

                if (je.isDirectory()) {
                    mev.setEntry(null, je);
                    return;
                }

                if (uname.endsWith(".DSA") || uname.endsWith(".RSA") ||
                    uname.endsWith(".SF")) {
                    /* We parse only DSA or RSA PKCS7 blocks. */
                    parsingBlockOrSF = true;
                    baos.reset();
                    mev.setEntry(null, je);
                }
                return;
            }
        }

        if (parsingMeta) {
            doneWithMeta();
        }

        if (je.isDirectory()) {
            mev.setEntry(null, je);
            return;
        }

        // be liberal in what you accept. If the name starts with ./, remove
        // it as we internally canonicalize it with out the ./.
        if (name.startsWith("./"))
            name = name.substring(2);

        // be liberal in what you accept. If the name starts with /, remove
        // it as we internally canonicalize it with out the /.
        if (name.startsWith("/"))
            name = name.substring(1);

        // only set the jev object for entries that have a signature
        if (sigFileCerts.get(name) != null) {
            mev.setEntry(name, je);
            return;
        }

        // don't compute the digest for this entry
        mev.setEntry(null, je);

        return;
    }

    /**
     * update a single byte.
     */

    public void update(int b, ManifestEntryVerifier mev)
        throws IOException
    {
        if (b != -1) {
            if (parsingBlockOrSF) {
                baos.write(b);
            } else {
                mev.update((byte)b);
            }
        } else {
            processEntry(mev);
        }
    }

    /**
     * update an array of bytes.
     */

    public void update(int n, byte[] b, int off, int len,
                       ManifestEntryVerifier mev)
        throws IOException
    {
        if (n != -1) {
            if (parsingBlockOrSF) {
                baos.write(b, off, n);
            } else {
                mev.update(b, off, n);
            }
        } else {
            processEntry(mev);
        }
    }

    /**
     * called when we reach the end of entry in one of the read() methods.
     */
    private void processEntry(ManifestEntryVerifier mev)
        throws IOException
    {
        if (!parsingBlockOrSF) {
            JarEntry je = mev.getEntry();
            if ((je != null) && (je.certs == null)) {
                je.certs = mev.verify(verifiedCerts, sigFileCerts);
            }
        } else {

            try {
                parsingBlockOrSF = false;

                if (debug != null) {
                    debug.println("processEntry: processing block");
                }

                String uname = mev.getEntry().getName()
                                             .toUpperCase(Locale.ENGLISH);

                if (uname.endsWith(".SF")) {
                    String key = uname.substring(0, uname.length()-3);
                    byte bytes[] = baos.toByteArray();
                    // add to sigFileData in case future blocks need it
                    sigFileData.put(key, bytes);
                    // check pending blocks, we can now process
                    // anyone waiting for this .SF file
                    Iterator it = pendingBlocks.iterator();
                    while (it.hasNext()) {
                        SignatureFileVerifier sfv =
                            (SignatureFileVerifier) it.next();
                        if (sfv.needSignatureFile(key)) {
                            if (debug != null) {
                                debug.println(
                                 "processEntry: processing pending block");
                            }

                            sfv.setSignatureFile(bytes);
                            sfv.process(sigFileCerts, manifestDigests);   /*ibm@160250*/
                        }
                    }
                    return;
                }

                // now we are parsing a signature block file

                String key = uname.substring(0, uname.lastIndexOf("."));

                if (certCache == null)
                    certCache = new ArrayList();

                if (manDig == null) {
                    synchronized(manifestRawBytes) {
                        if (manDig == null) {
                            manDig = new ManifestDigester(manifestRawBytes);
                            manifestRawBytes = null;
                        }
                    }
                }

                SignatureFileVerifier sfv =
                  new SignatureFileVerifier(certCache,
                                            manDig, uname, baos.toByteArray());

                if (sfv.needSignatureFileBytes()) {
                    // see if we have already parsed an external .SF file
                    byte[] bytes = (byte[]) sigFileData.get(key);

                    if (bytes == null) {
                        // put this block on queue for later processing
                        // since we don't have the .SF bytes yet
                        // (uname, block);
                        if (debug != null) {
                            debug.println("adding pending block");
                        }
                        pendingBlocks.add(sfv);
                        return;
                    } else {
                        sfv.setSignatureFile(bytes);
                    }
                }
                sfv.process(sigFileCerts, manifestDigests);      /*ibm@160250*/

            } catch (sun.security.pkcs.ParsingException pe) {
                if (debug != null) {
                    debug.println("processEntry caught: "+pe);         /*ibm@63626*/
                    debug.printStackTrace(pe);                         /*ibm@63626*/
                }
                // ignore and treat as unsigned
            } catch (IOException ioe) {
                if (debug != null) {
                    debug.println("processEntry caught: "+ioe);        /*ibm@63626*/
                    debug.printStackTrace(ioe);                        /*ibm@63626*/
                }
                // ignore and treat as unsigned
            } catch (SignatureException se) {
                if (debug != null) {
                    debug.println("processEntry caught: "+se);         /*ibm@63626*/
                    debug.printStackTrace(se);                         /*ibm@63626*/
                }
                // ignore and treat as unsigned
            } catch (NoSuchAlgorithmException nsae) {
                if (debug != null) {
                    debug.println("processEntry caught: "+nsae);       /*ibm@63626*/
                    debug.printStackTrace(nsae);                       /*ibm@63626*/
                }
                // ignore and treat as unsigned
            }
        }
    }

    /**
     * return an array of java.security.cert.Certificate objects for
     * the given file in the jar. this array is not cloned.
     * deprecated.
     */
    public java.security.cert.Certificate[] getCerts(String name)
    {
        return (java.security.cert.Certificate[])verifiedCerts.get(name);
    }

    /*ibm@160250-Starts*/
    /**
     * return an array of java.security.cert.Certificate objects for
     * the given file in the jar. this array is not cloned.
     *
     */
    public java.security.cert.Certificate[] getCerts(JarFile jar, JarEntry entry)
    {
        String name = entry.getName();
        if (eagerValidation && sigFileCerts.get(name) != null) {
            /*
             * Force a read of the entry data to generate the
             * verification hash.
             */
            try {
                InputStream s = jar.getInputStream(entry);
                byte[] buffer = new byte[1024];
                int n = buffer.length;
                while (n != -1) {
                    n = s.read(buffer, 0, buffer.length);
                }
                s.close();
            } catch (IOException e) {
            }
        }
        return getCerts(name);
    }
    /*ibm@160250-Ends*/

    /**
     * returns true if there no files to verify.
     * should only be called after all the META-INF entries
     * have been processed.
     */
    boolean nothingToVerify()
    {
        return (anyToVerify == false);
    }

    /**
     * called to let us know we have processed all the
     * META-INF entries, and if we re-read one of them, don't
     * re-process it. Also gets rid of any data structures
     * we needed when parsing META-INF entries.
     */
    void doneWithMeta()
    {
        parsingMeta = false;
        anyToVerify = !sigFileCerts.isEmpty();
        baos = null;
        sigFileData = null;
        pendingBlocks = null;
        certCache = null;
        manDig = null;
    }

    static class VerifierStream extends java.io.InputStream {

        private InputStream is;
        private JarVerifier jv;
        private ManifestEntryVerifier mev;
        private long numLeft;

        VerifierStream(Manifest man,
                       JarEntry je,
                       InputStream is,
                       JarVerifier jv) throws IOException
        {
            this.is = is;
            this.jv = jv;
            this.mev = new ManifestEntryVerifier(man);
            this.jv.beginEntry(je, mev);
            this.numLeft = je.getSize();
	    if (this.numLeft == 0)
		this.jv.update(-1, this.mev);
	}

        public int read() throws IOException
        {
            if (numLeft > 0) {
                int b = is.read();
                jv.update(b, mev);
                numLeft--;
                if (numLeft == 0)
                    jv.update(-1, mev);
                return b;
            } else {
                return -1;
            }
        }

        public int read(byte b[], int off, int len) throws IOException {
            if ((numLeft > 0) && (numLeft < len)) {
                len = (int)numLeft;
            }

            if (numLeft > 0) {
                int n = is.read(b, off, len);
                jv.update(n, b, off, len, mev);
                numLeft -= n;
                if (numLeft == 0)
                    jv.update(-1, b, off, len, mev);
                return n;
            } else {
                return -1;
            }
        }

        public void close()
            throws IOException
        {
            if (is != null)
                is.close();
            is = null;
            mev = null;
            jv = null;
        }

        public int available() throws IOException {
            return is.available();
        }

    }

    /*ibm@160250-Starts*/
    // Extended JavaUtilJarAccess CodeSource API Support

    private Map urlToCodeSourceMap = new HashMap();
    private Map certsToCodeSource = new HashMap();
    private URL lastURL;
    private Map lastURLMap;

    /*
     * Create a unique mapping from codeSigner cache entries to CodeSource.
     * In theory, multiple URLs origins could map to a single locally cached
     * and shared JAR file although in practice there will be a single URL in use.
     */
    private synchronized CodeSource mapCertsToCodeSource(URL url, Certificate[] certs) {
        Map map;
        if (url == lastURL) {
            map = lastURLMap;
        } else {
            map = (Map)urlToCodeSourceMap.get(url);
            if (map == null) {
                map = new HashMap();
                urlToCodeSourceMap.put(url, map);
            }
            lastURLMap = map;
            lastURL = url;
        }
        CodeSource cs = (CodeSource) map.get(certs);
        if (cs == null) {
            cs = new VerifierCodeSource(csdomain, url, certs);
            certsToCodeSource.put(certs, cs);
        }
        return cs;
    }

    private CodeSource[] mapCertsToCodeSources(URL url, List certs, boolean unsigned) {
        List sources = new ArrayList();

        for (int i=0; i < certs.size(); i++) {
            sources.add(mapCertsToCodeSource(url, (Certificate[]) certs.get(i)));
        }
        if (unsigned) {
            sources.add(mapCertsToCodeSource(url, null));
        }
        return (CodeSource[])sources.toArray(new CodeSource[sources.size()]);
    }

    private Certificate[] emptyCerts = new Certificate[0];

    /*
     * Match CodeSource to a Certificate[] in the signer cache.
     */
    private Certificate[] findMatchingCerts(CodeSource cs) {
        if (cs instanceof VerifierCodeSource) {
            VerifierCodeSource vcs = (VerifierCodeSource) cs;
            if (vcs.isSameDomain(csdomain)) {
                return ((VerifierCodeSource)cs).getPrivateCerts();
            }
        }

        /*
         * In practice signers should always be optimized above
         * but this handles a CodeSource of any type, just in case.
         */
        CodeSource[] sources = mapCertsToCodeSources(cs.getLocation(), getJarCerts(), true);
        List sourceList = new ArrayList();
        for (int i=0; i < sources.length; i++) {
            sourceList.add(sources[i]);
        }
        int j = sourceList.indexOf(cs);
        if (j != -1) {
            Certificate[] match;
            match = ((VerifierCodeSource)sourceList.get(j)).getPrivateCerts();
            if (match == null) {
                match = emptyCerts;
            }
            return match;
        }
        return null;
    }

    /*
     * Instances of this class hold uncopied references to internal
     * signing data that can be compared by object reference identity.
     */
    private static class VerifierCodeSource extends CodeSource {
        URL vlocation;
        java.security.cert.Certificate[] vcerts;
        Object csdomain;

        VerifierCodeSource(Object csdomain, URL location, java.security.cert.Certificate[] certs) {
            super(location, certs);
            this.csdomain = csdomain;
            vlocation = location;
            vcerts = certs; // from signerCache
        }

        /*
         * All VerifierCodeSource instances are constructed based on
         * singleton signerCache or signerCacheCert entries for each unique signer.
         * No CodeSigner<->Certificate[] conversion is required.
         * We use these assumptions to optimize equality comparisons.
         */
        public boolean equals(Object obj) {
            if (obj == this) {
                return true;
            }
            if (obj instanceof VerifierCodeSource) {
                VerifierCodeSource that = (VerifierCodeSource) obj;

                /*
                 * Only compare against other per-signer singletons constructed
                 * on behalf of the same JarFile instance. Otherwise, compare
                 * things the slower way.
                 */
                if (isSameDomain(that.csdomain)) {
                    if (that.vcerts != this.vcerts) {
                        return false;
                    }
                    if (that.vlocation != null) {
                        return that.vlocation.equals(this.vlocation);
                    } else if (this.vlocation != null) {
                        return this.vlocation.equals(that.vlocation);
                    } else { // both null
                        return true;
                    }
                }
            }
            return super.equals(obj);
        }

        boolean isSameDomain(Object csdomain) {
            return this.csdomain == csdomain;
        }

        private java.security.cert.Certificate[] getPrivateCerts() {
            return vcerts;
        }
    }

    private Map certsMap;

    private synchronized Map certsMap() {
        if (certsMap == null) {
            /*
             * Snapshot signer state so it doesn't change on us. We care
             * only about the asserted signatures. Verification of
             * signature validity happens via the JarEntry apis.
             */
            certsMap = new HashMap(verifiedCerts.size() + sigFileCerts.size());
            certsMap.putAll(verifiedCerts);
            certsMap.putAll(sigFileCerts);
        }
        return certsMap;
    }

    public synchronized Enumeration entryNames(JarFile jar, final CodeSource[] cs) {
        final Map map = certsMap();
        final Iterator itor = map.entrySet().iterator();
        boolean matchUnsigned = false;

        /*
         * Grab a single copy of the Certificate arrays. Check
         * to see if we can optimize Certificate equality test.
         */
        List req = new ArrayList(cs.length);
        for (int i=0; i < cs.length; i++) {
            Certificate[] match = findMatchingCerts(cs[i]);
            if (match != null) {
                if (match.length > 0) {
                    req.add(match);
                } else {
                    matchUnsigned = true;
                }
            }
        }

        final List certsReq = req;
        final Enumeration enum2 = (matchUnsigned) ? unsignedEntryNames(jar) : emptyEnumeration;

        return new Enumeration() {
            String name;

            public boolean hasMoreElements() {
                if (name != null) {
                    return true;
                }

                while (itor.hasNext()) {
                    Map.Entry e = (Map.Entry)itor.next();
                    if (certsReq.contains((Certificate[])e.getValue())) {
                        name = (String)e.getKey();
                        return true;
                    }
                }
                while (enum2.hasMoreElements()) {
                    name = (String)enum2.nextElement();
                    return true;
                }
                return false;
            }
            public Object nextElement() {
                if (hasMoreElements()) {
                    String value = name;
                    name = null;
                    return value;
                }
                throw new NoSuchElementException();
            }
        };
    }

    /*
     * Like entries() but screens out internal JAR mechanism entries
     * and includes signed entries with no ZIP data.
     */
    public Enumeration entries2(final JarFile jar, Enumeration e) {
        final Map map = new HashMap();
        map.putAll(certsMap());
        final Enumeration enum_ = e;
        return new Enumeration() {
            Enumeration signers = null;
            JarEntry entry;
            public boolean hasMoreElements() {
                if (entry != null) {
                    return true;
                }
                while (enum_.hasMoreElements()) {
                    ZipEntry ze = (ZipEntry) enum_.nextElement();
                    if (JarVerifier.isSigningRelated(ze.getName())) {
                        continue;
                    }
                    entry = jar.newEntry(ze);
                    return true;
                }
                if (signers == null) {
                    signers = Collections.enumeration(map.keySet());
                }
                while (signers.hasMoreElements()) {
                    String name = (String)signers.nextElement();
                    entry = jar.newEntry(new ZipEntry(name));
                    return true;
                }

                // Any map entries left?
                return false;
            }
            public Object nextElement() {
                if (hasMoreElements()) {
                     JarEntry je = entry;
                     map.remove(je.getName());
                     entry = null;
                     return je;
                 }
                 throw new NoSuchElementException();
            }
        };
    }

    private Enumeration emptyEnumeration = new Enumeration() {
        public boolean hasMoreElements() {
            return false;
        }
        public Object nextElement() {
            throw new NoSuchElementException();
        }
    };

    // true if file is part of the signature mechanism itself
    static boolean isSigningRelated(String name) {
        name = name.toUpperCase(Locale.ENGLISH);
        if (!name.startsWith("META-INF/")) {
            return false;
        }
        name = name.substring(9);
        if (name.indexOf('/') != -1) {
            return false;
        }
        if (name.endsWith(".DSA") ||
            name.endsWith(".RSA") ||
            name.endsWith(".SF")  ||
            name.startsWith("SIG-") ||
            name.equals("MANIFEST.MF")) {
            return true;
        }
        return false;
    }

    private Enumeration unsignedEntryNames(JarFile jar) {
        final Map map = certsMap();
        final Enumeration entries = jar.entries();
        return new Enumeration() {
            String name;

            /*
             * Grab entries from ZIP directory but screen out
             * metadata.
             */
            public boolean hasMoreElements() {
                if (name != null) {
                    return true;
                }
                while (entries.hasMoreElements()) {
                    String value;
                    ZipEntry e = (ZipEntry) entries.nextElement();
                    value = e.getName();
                    if (e.isDirectory() || isSigningRelated(value)) {
                        continue;
                    }
                    if (map.get(value) == null) {
                        name = value;
                        return true;
                    }
                }
                return false;
            }
            public Object nextElement() {
                if (hasMoreElements()) {
                     String value = name;
                     name = null;
                     return value;
                 }
                 throw new NoSuchElementException();
            }
        };
    }

    private List jarCerts;

    private synchronized List getJarCerts() {
        Certificate[] certs;
        if (jarCerts == null) {
            HashSet set = new HashSet();
            set.addAll(certsMap().values());
            jarCerts = new ArrayList();
            jarCerts.addAll(set);
        }
        return jarCerts;
    }

    public synchronized CodeSource[] getCodeSources(JarFile jar, URL url) {
        boolean hasUnsigned = unsignedEntryNames(jar).hasMoreElements();

        return mapCertsToCodeSources(url, getJarCerts(), hasUnsigned);
    }

    public CodeSource getCodeSource(URL url, String name) {
        Certificate[] certs;

        certs = (Certificate[]) certsMap().get(name);
        return mapCertsToCodeSource(url, certs);
    }

    public CodeSource getCodeSource(URL url, JarFile jar, JarEntry je) {
        Certificate[] certs;

        return mapCertsToCodeSource(url, getCerts(jar, je));
    }

    public void setEagerValidation(boolean eager) {
        eagerValidation = eager;
    }

    public synchronized List getManifestDigests() {
        return Collections.unmodifiableList(manifestDigests);
    }

    static CodeSource getUnsignedCS(URL url) {
        return new VerifierCodeSource(null, url, (java.security.cert.Certificate[]) null);
    }
    /*ibm@160250-Ends*/
}
